@@ -12,7 +12,7 @@ use File::Basename ();
 use URI::Escape ();
 use Digest::MD5;
 
-our $VERSION = '0.990116';
+our $VERSION = '0.990117';
 
 use constant CHUNK_SIZE => 4096;
 
@@ -24,7 +24,7 @@ BEGIN {
             last;
         }
     }
-    *_DEFAULT_CACHE_DIR = sub () { $cache_dir };
+    use constant _DEFAULT_CACHE_DIR => $cache_dir;
 }
 
 my $ESCAPE = {
@@ -80,6 +80,7 @@ sub new {
     $attrs->{encoding}     = 'utf-8';
     $attrs->{escape_html}  = 1;
     $attrs->{helpers}      = {};
+    $attrs->{helpers_options} = {};
     $attrs->{format}       = 'xhtml';
     $attrs->{prepend}      = '';
     $attrs->{append}       = '';
@@ -142,6 +143,7 @@ sub escape_html   { @_ > 1 ? $_[0]->{escape_html}   = $_[1] : $_[0]->{escape_htm
 sub code          { @_ > 1 ? $_[0]->{code}          = $_[1] : $_[0]->{code} }
 sub compiled      { @_ > 1 ? $_[0]->{compiled}      = $_[1] : $_[0]->{compiled} }
 sub helpers       { @_ > 1 ? $_[0]->{helpers}       = $_[1] : $_[0]->{helpers} }
+sub helpers_options { @_ > 1 ? $_[0]->{helpers_options} = $_[1] : $_[0]->{helpers_options} }
 sub filters       { @_ > 1 ? $_[0]->{filters}       = $_[1] : $_[0]->{filters} }
 sub prepend       { @_ > 1 ? $_[0]->{prepend}       = $_[1] : $_[0]->{prepend} }
 sub append        { @_ > 1 ? $_[0]->{append}        = $_[1] : $_[0]->{append} }
@@ -170,9 +172,10 @@ our @AUTOCLOSE = (qw/meta img link br hr input area param col base/);
 
 sub add_helper {
     my $self = shift;
-    my ($name, $code) = @_;
+    my ($name, $code, %options) = @_;
 
     $self->helpers->{$name} = $code;
+    $self->helpers_options->{$name} = \%options;
 }
 
 sub add_filter {
@@ -240,6 +243,7 @@ sub parse {
 
     my $level;
     my @multiline_el_queue;
+    my $multiline_code_el = undef;
     my @lines = split /\n/, $tmpl;
     push @lines, '' if $tmpl =~ m/\n$/;
     @lines = ('') if $tmpl eq "\n";
@@ -255,6 +259,11 @@ sub parse {
 
         my $el = {level => $level, type => 'text', line => $line, lineno => $i+1};
 
+        if (defined $multiline_code_el && $line =~ /^[-!=%#.:]/) {
+            push @$tape, $multiline_code_el;
+            undef $multiline_code_el;
+        }
+
         # Haml comment
         if ($line =~ m/^$comment_token(?: (.*))?/) {
             $el->{type} = 'comment';
@@ -311,12 +320,25 @@ sub parse {
             next;
         }
 
-        # Block
-        if ($line =~ s/^- \s*(.*)//) {
+        # Block (note even the final multiline block must end in |)
+        if ($line =~ s/^- \s*(.*)(\s\|\s*)$// ||
+            $line =~ s/^- \s*(.*)// ||
+                (defined $multiline_code_el && $line =~ s/^(.*)(\s\|\s*)$//)) {
+
             $el->{type} = 'block';
+            
+            if ($2) {
+                $multiline_code_el ||= $el;
+                $multiline_code_el->{text} ||= '';
+                $multiline_code_el->{text} .= $1;
+
+                next;
+            }
+            
             $el->{text} = $1;
             push @$tape, $el;
             next;
+            
         }
 
         # Preserve whitespace
@@ -333,8 +355,6 @@ sub parse {
         if ($line =~ m/^(?:$tag_start
             |$class_start
             |$id_start
-            |$attributes_start[^$attributes_start]
-            |$attributes_start2
             )/x
           )
         {
@@ -530,6 +550,24 @@ sub _update_lineno {
     return;
 }
 
+sub _open_implicit_brace {
+    my ($lines) = @_;
+        if (scalar(@$lines) && $lines->[-1] eq '}') {
+        pop @$lines;
+    } else {
+        push @$lines, '{';
+    }
+}
+
+sub _close_implicit_brace {
+    my ($lines) = @_;
+    if (scalar(@$lines) && $lines->[-1] eq '{') {
+        pop @$lines;
+    } else {
+        push @$lines, '}';
+    }
+}
+
 sub build {
     my $self = shift;
     my %vars = @_;
@@ -548,24 +586,49 @@ EOF
 
     $ESCAPE =~ s/\n//g;
 
-    my $namespace = $self->namespace || ref($self) . '::template';
+    # ensure namespace is set so that (for now) helpers
+    # can access outs & outs_raw (until we correctly allow
+    # helpers in `=` lines to capture their blocks eg. for `surrounds`
+    
+    if (! $self->namespace) {
+        $self->namespace(ref($self) . '::template');
+    }
+
+    my $namespace = $self->namespace;
     $code .= qq/package $namespace;/;
 
-    $code .= qq/sub { my \$_H = ''; $ESCAPE;/;
+    $code .= qq/sub { my \$_H = ''; $ESCAPE; /;
 
     $code .= qq/my \$self = shift;/;
+    $code .= qq/\$${namespace}::__self = \$self;/;
 
     $code .= qq/my \%____vars = \@_;/;
 
     $code .= qq/no strict 'refs'; no warnings 'redefine';/;
 
+    # using [1] since when called with arrow from namespace, [0] will be the namespace
+    $code .= qq/*${namespace}::outs = sub { \$_H .= escape(\$_[1]) };/;
+    $code .= qq/*${namespace}::outs_raw = sub { \$_H .= \$_[1] };/;
+    $code .= qq/*${namespace}::out_chomp = sub { chomp \$_H };/;
+
     # Install helpers
     for my $name (sort keys %{$self->helpers}) {
         next unless $name =~ m/^\w+$/;
 
-        $code .= "sub $name;";
-        $code .= " *$name = sub { \$self";
-        $code .= "->helpers->{'$name'}->(\$self->helpers_arg, \@_) };";
+        my $options = $self->{helpers_options}{$name} || {};
+
+        # allow bareword helpers and block capturing with optional helper prototypes
+        my $prototype = $options->{prototype};
+        $prototype = defined $prototype ? "($prototype)" : '';
+
+        # this option allows per-helper overriding of the helper_arg, important for builtin
+        # helpers to be safe in assuming the arg is self
+        my $helper_arg_code = $options->{arg_force_self} ? "\$${namespace}::__self" : "\$${namespace}::__self->helpers_arg";
+
+        # sub must be defined inside BEGIN {} for the prototype to be ready before main helper code is
+        # compiled
+        $code .= "BEGIN { \*${namespace}::${name} = sub $prototype { ";
+        $code .= "\$${namespace}::__self->helpers->{'$name'}->($helper_arg_code, \@_) }; } ";
     }
 
     # Install variables
@@ -597,7 +660,7 @@ EOF
         $level -= 2 * $in_block if $in_block;
 
         my $offset = '';
-        $offset .= ' ' x $level;
+        $offset .= ' ' x $level if $level > 0;
 
         my $escape = '';
         if (   (!exists $el->{escape} && $self->escape_html)
@@ -617,6 +680,7 @@ EOF
             {
                 pop @$stack;
                 undef $prev_stack_el;
+                _close_implicit_brace(\@lines);
             }
             else {
                 next ELEM;
@@ -631,7 +695,7 @@ EOF
             while (my $poped = pop @$stack) {
                 my $level = $poped->{level};
                 $level -= 2 * $in_block if $in_block;
-                my $poped_offset = ' ' x $level;
+                my $poped_offset = $level > 0 ? ' ' x $level : '';
 
                 my $ending = '';
                 if ($poped->{type} eq 'tag') {
@@ -646,6 +710,12 @@ EOF
                     push @lines, qq|\$_H .= "$poped_offset$ending\n";|;
                 }
 
+                _close_implicit_brace(\@lines);
+                
+                if ($poped->{type} eq 'block') {
+                    _close_implicit_brace(\@lines);
+                }
+
                 last STACKEDBLK if $poped->{level} == $el->{level};
             }
         }
@@ -749,6 +819,7 @@ EOF
                 }
                 elsif (!$el->{autoclose}) {
                     push @$stack, $el;
+                    _open_implicit_brace(\@lines);
                 }
 
                 $output .= qq|. "\n"|;
@@ -778,8 +849,10 @@ EOF
             }
 
             if ($el->{type} eq 'block') {
-                push @lines,  $el->{text};
+                _open_implicit_brace(\@lines);
+                push @lines,  ';' . $el->{text};
                 push @$stack, $el;
+                _open_implicit_brace(\@lines);
 
                 if ($prev_el && $prev_el->{level} > $el->{level}) {
                     $in_block--;
@@ -804,6 +877,7 @@ EOF
                 else {
                     $output .= qq/. "\n"/;
                     push @$stack, $el;
+                    _open_implicit_brace(\@lines);
                 }
 
                 $output .= qq/;/;
@@ -812,6 +886,7 @@ EOF
 
             if ($el->{type} eq 'comment') {
                 push @$stack, $el;
+                _open_implicit_brace(\@lines);
                 last SWITCH;
             }
 
@@ -844,7 +919,14 @@ EOF
         }    #end:SWITCH
     }    #end:ELEM
     continue {
+
+        # by bracing the content blocks, we will continue any existing block at the same level.
+        # this is important eg. if previously at this level the template has declared a `my`
+        # variable.
+        
+        _open_implicit_brace(\@lines);
         push @lines, '$_H .= ' . $output if $output;
+        _close_implicit_brace(\@lines);
         $output = '';
         $count++;
     }    #ELEM
@@ -853,7 +935,7 @@ EOF
     $last_empty_line = 1
       if $self->tape->[-1] && $self->tape->[-1]->{line} eq '';
 
-    # Close remaining content blocks, last-seen first
+    # Close remaining conten tblocks, last-seen first
     foreach my $el (reverse @$stack) {
         my $offset = ' ' x $el->{level};
         my $ending = '';
@@ -866,10 +948,21 @@ EOF
         }
 
         push @lines, qq|\$_H .= "$offset$ending\n";| if $ending;
+
+        _close_implicit_brace(\@lines);
+        if ($el->{type} eq 'block') {
+            _close_implicit_brace(\@lines);
+        }
+
     }
 
-    if ($lines[-1]) {
-        $lines[-1] =~ s/\n";$/";/ unless $last_empty_line;
+    if ($lines[-1] && !$last_empty_line) {
+        # usually (always?) there will be a closing '}' after the actual last .=
+        if ($lines[-2] && $lines[-1] eq '}') {
+            $lines[-2] =~ s/\n";$/";/;
+        } else {
+            $lines[-1] =~ s/\n";$/";/;
+        }
     }
 
     $code .= join("\n", @lines);
@@ -965,7 +1058,7 @@ sub _parse_interpolation {
 
     return '' unless @parts;
 
-    return '" . ' . join('.', map {s/\\\\#\\{/#\\{/; $_} @parts) . '."';
+    return '" . ' . join('.', map {s/\\\\#\\\{/#\\\{/; $_} @parts) . '."';
 }
 
 sub compile {
@@ -1570,6 +1663,8 @@ Jason Younker
 
 TheAthlete
 
+Mark Aufflick (aufflick)
+
 =head1 COPYRIGHT AND LICENSE
 
 Copyright (C) 2009-2013, Viacheslav Tykhanovskyi.
